/*
 *  Copyright 2008-2011 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

#include <thrust/iterator/iterator_traits.h>
#include <thrust/detail/device/dereference.h>

#include <stdio.h>

#include <vector>
#include <numeric>
#include "omp.h"

namespace thrust
{
namespace detail
{
namespace device
{
namespace omp
{

template<typename InputIterator,
         typename OutputIterator,
         typename AssociativeOperator>
  OutputIterator inclusive_scan(InputIterator first,
                                InputIterator last,
                                OutputIterator result,
                                AssociativeOperator binary_op)
{
  typedef typename thrust::iterator_traits<OutputIterator>::value_type OutputType;

  // If there is only one processor or one or fewer data elements, don't do extra work
  int numThreads = omp_get_max_threads();
  int N = last - first;
  if (N <= 0) return (result);
  if (N == 1) { result[0] = first[0]; return (result + N); } 
  if (numThreads < 2)
  {
    return std::partial_sum(first, last, result, binary_op);
  }

  // Initialize variables for data size and how many elements each processor gets
  if (numThreads > N) numThreads = N;
  int itemsPerThread = N / numThreads;

  // Each processor sums (with respect to binary scan operator) all elements assigned to it
  std::vector<typename std::iterator_traits<InputIterator>::value_type> processorSums(numThreads); 
  std::vector<typename std::iterator_traits<InputIterator>::value_type> processorSuppl(numThreads-1);
  std::vector<bool> processorSupplValid(numThreads-1); 
  std::fill(processorSupplValid.begin(), processorSupplValid.end(), false);
  int sumItemsPerThread = std::max(1, ((numThreads-1)*itemsPerThread)/numThreads);
  #pragma omp parallel
  {
    int id = omp_get_thread_num();
    if (id < numThreads - 1)
    {
      int startIndex = id*itemsPerThread;
      int endIndex = startIndex + sumItemsPerThread;
      typename std::iterator_traits<InputIterator>::value_type sum = first[startIndex];
      for (int i=startIndex+1; i<endIndex; i++)
        sum = binary_op(sum, first[i]);
      processorSums[id] = sum;
    }
    else if (id == numThreads - 1)
    {
      for (int s=0; s<numThreads-1; s++)
      {
        int startIndex = s*itemsPerThread + sumItemsPerThread;
        int endIndex = (s+1)*itemsPerThread;
        typename std::iterator_traits<InputIterator>::value_type psum = first[startIndex];
        for (int i=startIndex+1; i<endIndex; i++)
          psum = binary_op(psum, first[i]);
        if (endIndex > startIndex)
        { 
          processorSuppl[s] = psum;
          processorSupplValid[s] = true;
        }
      }
    }
  }
  
  for (int i=0; i<numThreads-1; i++)
    if (processorSupplValid[i])
      processorSums[i] = binary_op(processorSums[i], processorSuppl[i]);
  
  // Perform a scan across the processor sums to get offsets for each processor
  typename std::iterator_traits<InputIterator>::value_type lastItem = processorSums[0];
  typename std::iterator_traits<InputIterator>::value_type newLastItem = processorSums[1];
  processorSums[1] = lastItem;
  lastItem = newLastItem;
  for (int i=2; i<numThreads; i++)
  {
    typename std::iterator_traits<InputIterator>::value_type newLastItem = processorSums[i]; 
    processorSums[i] = binary_op(processorSums[i-1], lastItem);
    lastItem = newLastItem;
  }
 
  // Each processor scans the elements assigned to it, using result of processor scan above as offset
  #pragma omp parallel
  {
    int id = omp_get_thread_num();
    if (id < numThreads)
    {
      typename std::iterator_traits<InputIterator>::value_type lastItem = first[id*itemsPerThread]; 
      if (id > 0) result[id*itemsPerThread] = binary_op(processorSums[id], lastItem);
      else result[id*itemsPerThread] = lastItem;
      int firstIndex = id*itemsPerThread+1;
      int lastIndex = (id+1)*itemsPerThread;
      if (id == numThreads-1) lastIndex = N;

      typename std::iterator_traits<InputIterator>::value_type newLastItem;
      typename std::iterator_traits<InputIterator>::value_type prevResult = result[firstIndex-1];
      
      for (int i=firstIndex; i<lastIndex; i++)
        prevResult = result[i] = binary_op(prevResult, first[i]);    
    }
  }
  return (result + N);
}


template<typename InputIterator,
         typename OutputIterator,
         typename T,
         typename AssociativeOperator>
  OutputIterator exclusive_scan(InputIterator first,
                                InputIterator last,
                                OutputIterator result,
                                T init,
                                AssociativeOperator binary_op)
{
  typedef typename thrust::iterator_traits<OutputIterator>::value_type OutputType;
  OutputType initOut = init;

  // If there is only one processor or one or fewer data elements, don't do extra work
  int numThreads = omp_get_max_threads();
  int N = last - first;
  if (N <= 0) return (result);
  if (N == 1) { result[0] = initOut; return (result + N); } 
  if (numThreads < 2)
  {
    typename std::iterator_traits<InputIterator>::value_type lastItem = first[0];
    result[0] = initOut;
    typename std::iterator_traits<InputIterator>::value_type newLastItem;
    typename std::iterator_traits<InputIterator>::value_type prevResult = 0;
    for (int i=1; i<N; i++)
    {
      newLastItem = first[i];
      prevResult = result[i] = binary_op(prevResult, lastItem);
      lastItem = newLastItem;
    }
    return (result + N);
  }

  // Initialize variables for data size and how many elements each processor gets
  if (numThreads > N) numThreads = N;
  int itemsPerThread = N / numThreads;

  // Each processor sums (with respect to binary scan operator) all elements assigned to it
  std::vector<typename std::iterator_traits<InputIterator>::value_type> processorSums(numThreads); 
  std::vector<typename std::iterator_traits<InputIterator>::value_type> processorSuppl(numThreads-1);
  std::vector<bool> processorSupplValid(numThreads-1); 
  std::fill(processorSupplValid.begin(), processorSupplValid.end(), false);
  int sumItemsPerThread = std::max(1, ((numThreads-1)*itemsPerThread)/numThreads);
  #pragma omp parallel
  {
    int id = omp_get_thread_num();
    if (id < numThreads - 1)
    {
      int startIndex = id*itemsPerThread;
      int endIndex = startIndex + sumItemsPerThread;
      typename std::iterator_traits<InputIterator>::value_type sum = first[startIndex];
      for (int i=startIndex+1; i<endIndex; i++)
        sum = binary_op(sum, first[i]);
      processorSums[id] = sum;
    }
    else if (id == numThreads - 1)
    {
      for (int s=0; s<numThreads-1; s++)
      {
        int startIndex = s*itemsPerThread + sumItemsPerThread;
        int endIndex = (s+1)*itemsPerThread;
        typename std::iterator_traits<InputIterator>::value_type psum = first[startIndex];
        for (int i=startIndex+1; i<endIndex; i++)
          psum = binary_op(psum, first[i]);
        if (endIndex > startIndex)
        { 
          processorSuppl[s] = psum;
          processorSupplValid[s] = true;
        }
      }
    }
  }
  
  for (int i=0; i<numThreads-1; i++)
    if (processorSupplValid[i])
      processorSums[i] = binary_op(processorSums[i], processorSuppl[i]);
 
  // Perform a scan across the processor sums to get offsets for each processor
  typename std::iterator_traits<InputIterator>::value_type lastItem = processorSums[0];
  typename std::iterator_traits<InputIterator>::value_type newLastItem = processorSums[1];
  processorSums[0] = init;
  processorSums[1] = lastItem;
  lastItem = newLastItem;
  for (int i=2; i<numThreads; i++)
  {
    typename std::iterator_traits<InputIterator>::value_type newLastItem = processorSums[i]; 
    processorSums[i] = binary_op(processorSums[i-1], lastItem);
    lastItem = newLastItem;
  }

  // Each processor scans the elements assigned to it, using result of processor scan above as offset
  #pragma omp parallel
  {
    int id = omp_get_thread_num();
    if (id < numThreads)
    {
      typename std::iterator_traits<InputIterator>::value_type lastItem = first[id*itemsPerThread]; 
      result[id*itemsPerThread] = processorSums[id];
      int firstIndex = id*itemsPerThread+1;
      int lastIndex = (id+1)*itemsPerThread;
      if (id == numThreads-1) lastIndex = N;

      typename std::iterator_traits<InputIterator>::value_type newLastItem;
      typename std::iterator_traits<InputIterator>::value_type prevResult = result[firstIndex-1];
      
      for (int i=firstIndex; i<lastIndex; i++)
      {
        newLastItem = first[i]; 
        prevResult = result[i] = binary_op(prevResult, lastItem); 
        lastItem = newLastItem;
      }     
    }
  }
  return (result + N);
}

} // end namespace omp
} // end namespace device
} // end namespace detail
} // end namespace thrust

